#include "b-spline.h"
#include <stdio.h>

struct point_struc {
  float x;
  float y;
};

struct b_spline_struc {
  Curve_B_spline object;
};

extern "C" struct b_spline_struc *b_spline_new()
{
  b_spline_struc *result = NULL;
  result = new b_spline_struc;
  return result;
}

extern "C" bool b_spline_add_control_poly_point
(struct b_spline_struc *b_spline, struct point_struc point)
{
  Vector<gtype> p(2);
  p.setItem(0, point.x);
  p.setItem(1, point.y);
  return b_spline->object.addControlPolyPoint(p);
}

extern "C" bool b_spline_remove_control_poly_point
(struct b_spline_struc *b_spline, struct point_struc point)
{
  Vector<gtype> p(2);
  p.setItem(0, point.x);
  p.setItem(1, point.y);
  return b_spline->object.removeControlPolyPoint(p);
}

extern "C" bool b_spline_update_control_poly_point
(struct b_spline_struc *b_spline, struct point_struc point,
 struct point_struc new_point)
{
  Vector<gtype> p(2), new_p(2);
  p.setItem(0, point.x);
  p.setItem(1, point.y);
  new_p.setItem(0, new_point.x);
  new_p.setItem(1, new_point.y);
  return b_spline->object.updateControlPolyPoint(p, new_p);
}

extern "C" int b_spline_get_control_poly_points_count
(struct b_spline_struc *b_spline)
{
  return b_spline->object.getControlPolyPointsCount();
}

extern "C" struct point_struc b_spline_get_control_poly_point
(struct b_spline_struc *b_spline, int i)
{
  Vector<gtype> retval = b_spline->object.getControlPolyPoint(i);
  point_struc result;
  result.x = retval.getItem(0);
  result.y = retval.getItem(1);
  return result;
}

extern "C" void b_spline_get_control_poly_points
(struct b_spline_struc *b_spline, int *n, struct point_struc **points)
{
  Vector<gtype> *retval;
  b_spline->object.getControlPolyPoints(*n, retval);
  if (*n == 0) return;
  *points = NULL;
  *points = new point_struc [*n];
  if (*points == NULL) return;
  for (int i = 0; i < *n; ++i) { 
    (*points)[i].x = retval[i].getItem(0);
    (*points)[i].y = retval[i].getItem(1);
  }
}

extern "C" bool b_spline_set_level(struct b_spline_struc *b_spline, int l)
{
  return b_spline->object.setLevel(l);
}

extern "C" int b_spline_get_level(struct b_spline_struc *b_spline)
{
  return b_spline->object.getLevel();
}

extern "C" void b_spline_get_param_vector
(struct b_spline_struc *b_spline, int *n, float **par_vector)
{
  Vector<gtype> retval = b_spline->object.getParamVector();
  *n = retval.getDim();
  *par_vector = NULL;
  if (retval.getDim() == 0) return;
  *par_vector = new float [*n];
  if (*par_vector == NULL) return;
  for (int i = 0; i < *n; ++i) (*par_vector)[i] = retval.getItem(i);
}

extern "C" bool b_spline_set_param_vector
(struct b_spline_struc *b_spline, int n, float *par_vector)
{
  if (n < 0) return false;
  if (par_vector == NULL) return false;  
  Vector<gtype> pv (n);
  for (int i = 0; i < n; ++i) pv.setItem(i, par_vector[i]);
  return b_spline->object.setParamVector(pv);
}

extern "C" float b_spline_deBoor_base_func
(struct b_spline_struc *b_spline, int i, float u)
{
  return b_spline->object.deBoorBaseFunc(i, u);
}

extern "C" struct point_struc b_spline_deBoor_alg
(struct b_spline_struc *b_spline, float u)
{
  Vector<gtype> retval = b_spline->object.deBoorAlg(u);
  point_struc result;
  if (retval.getDim() == 0) return result;
  result.x = retval.getItem(0);
  result.y = retval.getItem(1);
  return result;
}
