#include "b-spline.h"

Curve_B_spline::Curve_B_spline ()
{
  control_poly_points_count = 0;
  control_poly_points = NULL;
  level = 0;
}

Curve_B_spline::~Curve_B_spline ()
{
  control_poly_points_count = 0;
  delete [] control_poly_points; control_poly_points = NULL;
  level = 0;
}

bool Curve_B_spline::addControlPolyPoint(Vector<gtype> point)
{
  if (control_poly_points_count != 0) 
    if (control_poly_points[0].getDim() != point.getDim()) return false;
  int i;
  Vector<gtype> *control_poly_points_new = NULL;
  control_poly_points_new = 
    new Vector<gtype> [control_poly_points_count + 1];
  if (control_poly_points_new == NULL) return false;
  for (i = 0; i < control_poly_points_count; ++i) 
    control_poly_points_new[i] = control_poly_points[i];
  control_poly_points_new[i] = point;
  ++control_poly_points_count;
  delete [] control_poly_points;
  control_poly_points = control_poly_points_new;
  setLevel(level);
  return true;
}

bool Curve_B_spline::removeControlPolyPoint(Vector<gtype> point)
{
  if (control_poly_points_count == 0) return false;
  else if (control_poly_points[0].getDim() != point.getDim()) return false;
  int i, j = -1;
  for (i = 0; i < control_poly_points_count; ++i)
    if (control_poly_points[i] == point) { j = i; break; }
  if (j == -1) return false;
  Vector<gtype> *control_poly_points_new = NULL;
  if (control_poly_points_count != 1) {
    control_poly_points_new = 
      new Vector<gtype> [control_poly_points_count - 1];
    if (control_poly_points_new == NULL) return false;
  }
  int k;
  for (i = 0, k = 0; i < control_poly_points_count; ++i) {
    if (i == j) continue;
    control_poly_points_new[k] = control_poly_points[i];
    ++k;
  }
  --control_poly_points_count;
  delete [] control_poly_points;
  control_poly_points = control_poly_points_new;
  setLevel(level);
  return true;
}

bool Curve_B_spline::updateControlPolyPoint
(Vector<gtype> point, Vector<gtype> new_point)
{
  if (control_poly_points_count == 0) return false;
  else if (control_poly_points[0].getDim() != point.getDim()) return false;
  int i, j = -1;
  for (i = 0; i < control_poly_points_count; ++i)
    if (control_poly_points[i] == point) { j = i; break; }
  if (j == -1) return false;
  for (i = 0; i < point.getDim(); ++i)
    control_poly_points[j].setItem(i, new_point.getItem(i));
  return true;
}

Vector<gtype> Curve_B_spline::getControlPolyPoint(int i)
{
  if ((i >= 0) && (i < control_poly_points_count))
    return control_poly_points[i];
  return Vector<gtype>();
}

void Curve_B_spline::getControlPolyPoints(int &n, Vector<gtype> *&points)
{
  points = NULL;
  n = control_poly_points_count;
  if (control_poly_points == 0) return;
  points = new Vector<gtype> [control_poly_points_count];
  if (points == NULL) return;
  for (int i = 0; i < control_poly_points_count; ++i)
    points[i] = control_poly_points[i];
}

bool Curve_B_spline::setLevel(int l)
{
  if ((l < 0) || (l >= control_poly_points_count)) return false;
  Vector<float> par_vector (control_poly_points_count-1+l+1+1);
  int less = param_vector.getDim();
  if (par_vector.getDim() < param_vector.getDim()) less = par_vector.getDim();
  int i;
  for (i = 0; i < less; ++i) par_vector.setItem(i, param_vector.getItem(i));
  if (i == 0) { par_vector.setItem(0, 0); ++i; }
  for (; i < par_vector.getDim(); ++i) 
    par_vector.setItem(i, par_vector.getItem(i-1) + 1);
  param_vector = par_vector;
  level = l;
  return true;
}

bool Curve_B_spline::setParamVector(Vector<float> par_vector)
{
  if (param_vector.getDim() != par_vector.getDim()) return false;
  for (int i = 0; i < param_vector.getDim(); ++i)
    param_vector.setItem(i, par_vector.getItem(i));
}

gtype Curve_B_spline::deBoorBaseFunc(int i, float u)
{
  if (control_poly_points == 0) return 0;
  return _deBoorBaseFunc(level, i, u);
}

gtype Curve_B_spline::_deBoorBaseFunc(int k, int i, float u)
{
  if (k == 0) {
    if ((u >= param_vector.getItem(i)) && (u < param_vector.getItem(i+1)))
      return 1;
    else return 0;
  }
  else
    return (((u - param_vector.getItem(i)) / 
	     (param_vector.getItem(i+k) - param_vector.getItem(i))) * 
	    _deBoorBaseFunc(k-1, i, u)) + 
      (((param_vector.getItem(i+k+1) - u) / 
	(param_vector.getItem(i+k+1) - param_vector.getItem(i+1))) * 
       _deBoorBaseFunc(k-1, i+1, u));
}

Vector<gtype> Curve_B_spline::deBoorAlg(float u)
{
  int l = -1;
  for (int i = 0; i < param_vector.getDim() - 1; ++i)
    if ((u >= param_vector.getItem(i)) && (u <= param_vector.getItem(i+1))) {
      l = i; break;
    }
  if ((l >= level) && (l < control_poly_points_count))
    return _deBoorAlg(level, l, u);
  else return Vector<gtype>();
}

Vector<gtype> Curve_B_spline::_deBoorAlg(int j, int i, float u)
{
  if (j == 0) return control_poly_points[i];
  else {
    float alpha = (u - param_vector.getItem(i)) / 
      (param_vector.getItem(i+level-j+1) - param_vector.getItem(i));
    return (_deBoorAlg(j-1, i-1, u) * (1 - alpha)) + 
      (_deBoorAlg(j-1, i, u) * alpha);
  }
}

