#include "b-spline-interface.h"
#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>

struct b_spline_struc *b_spline;

/* Backing pixmap for drawing area */
static GdkPixmap *pixmap = NULL;
GdkCursorType cursor_type = GDK_TCROSS;
GtkWidget *stat_label;
GtkWidget *entry;

const int control_point_size = 6;
int zoom = 4;
struct point_struc grab_point;
int show_control_poly = 0;

static void draw_curve(GtkWidget *widget);
void draw_control_points(GtkWidget* widget);
void redraw_scene(GtkWidget *widget);
static void draw_control_poly(GtkWidget *widget);

/* Create a new backing pixmap of the appropriate size */
static gint configure_event(GtkWidget *widget, GdkEventConfigure *event)
{
  if (pixmap) gdk_pixmap_unref(pixmap);
  pixmap = gdk_pixmap_new(widget->window, widget->allocation.width,
			  widget->allocation.height, -1);
  redraw_scene(widget);
  
  return TRUE;
}

void project(int width, int height, struct point_struc *point)
{
  point->x = (width / 2) + (point->x * zoom);
  point->y = (height / 2) - (point->y * zoom);
}

void unproject(int width, int height, struct point_struc *point)
{
  point->x = (point->x - (width / 2)) / zoom;
  point->y = ((height / 2) - point->y) / zoom;
}

/* Redraw the screen from the backing pixmap */
static gint expose_event(GtkWidget *widget, GdkEventExpose *event)
{
  gdk_draw_pixmap(widget->window, 
		  widget->style->fg_gc[GTK_WIDGET_STATE(widget)], pixmap,
		  event->area.x, event->area.y, event->area.x, event->area.y,
		  event->area.width, event->area.height);

  return FALSE;
}

/* Clear the backing pixmap */
static void clear_pixmap(GtkWidget *widget)
{
  gdk_draw_rectangle (pixmap, widget->style->white_gc, TRUE, 0, 0,
		      widget->allocation.width, widget->allocation.height);
}

void get_param_vector(char *buffer)
{
  char num[10];
  float *par_vector;
  int n, i;
  
  strcpy(buffer, "");
  b_spline_get_param_vector(b_spline, &n, &par_vector);
  for (i = 0; i < n; ++i) {
    sprintf (num, "%.1f", par_vector[i]);
    strcat(buffer, num);
    if (i != n-1) strcat(buffer, " ");
  }
}

void set_param_vector(GtkWidget *widget)
{
  char num[10], *buffer;
  float par_vector[100], fnum;
  int n = 0, i, j;
  
  strcpy(num, "");
  buffer = gtk_entry_get_text(GTK_ENTRY(entry));
  for (i = 0, j = 0; buffer[i]; ++i)
    if (buffer[i] != ' ') { num[j] = buffer[i]; ++j; }
    else { 
      num[j] = 0; sscanf(num, "%f", &fnum); par_vector[n] = fnum; ++n; j = 0;
    }
  num[j] = 0; sscanf(num, "%f", &fnum); par_vector[n] = fnum; ++n;
  b_spline_set_param_vector(b_spline, n, par_vector);
  redraw_scene(widget);
}

/* Redraw the the scene */
void redraw_scene(GtkWidget *widget)
{
  GdkRectangle update_rect;
  char buffer[100];

  update_rect.x = 0;
  update_rect.y = 0;
  update_rect.width = widget->allocation.width;
  update_rect.height = widget->allocation.height;

  sprintf (buffer, "Points: %i, Level: %i", 
	   b_spline_get_control_poly_points_count(b_spline),
	   b_spline_get_level(b_spline));
  gtk_label_set_text(GTK_LABEL(stat_label), buffer);
  get_param_vector(buffer);
  gtk_entry_set_text(GTK_ENTRY(entry), buffer);
  clear_pixmap(widget);
  draw_control_points(widget);
  if (show_control_poly) draw_control_poly(widget);
  draw_curve(widget);
  gtk_widget_draw (widget, &update_rect);
}

/* Draw a point on the screen */
/*
static void draw_point(GtkWidget *widget, gdouble x, gdouble y)
{
  GdkRectangle update_rect;

  update_rect.x = 0;
  update_rect.y = 0;
  update_rect.width = widget->allocation.width;
  update_rect.height = widget->allocation.height;

  gdk_draw_point (pixmap, widget->style->black_gc, x, y);
  gtk_widget_draw (widget, &update_rect);
}
*/
int get_approx_point
(GtkWidget *widget, struct point_struc *point, int x, int y)
{
  int n, i, gotit = 0;
  struct point_struc *points; 

  b_spline_get_control_poly_points(b_spline, &n, &points);
  for (i = 0; i < n; ++i) {
    project(widget->allocation.width, widget->allocation.height, &points[i]);
    if ((points[i].x >= x - control_point_size / 2) && 
	(points[i].x <= x + control_point_size / 2) &&
	(points[i].y >= y - control_point_size / 2) &&
	(points[i].y <= y + control_point_size / 2)) {
      *point = points[i]; gotit = 1; break;
    }
  }
  return gotit;
}

static gint button_press_event(GtkWidget *widget, GdkEventButton *event )
{
  struct point_struc point;
  
  if (event->button == 1 && pixmap != NULL) {
    point.x = event->x;
    point.y = event->y;
    if (!get_approx_point(widget, &grab_point, point.x, point.y)) {
      unproject(widget->allocation.width, widget->allocation.height, &point);
      b_spline_add_control_poly_point(b_spline, point);
      redraw_scene(widget);
    }
  }
  else if (event->button == 3 && pixmap != NULL) {
    point.x = event->x;
    point.y = event->y;
    if (get_approx_point(widget, &point, point.x, point.y)) {
      unproject(widget->allocation.width, widget->allocation.height, &point);
      b_spline_remove_control_poly_point(b_spline, point);
      redraw_scene(widget);
    }
  }

  return TRUE;
}

static gint motion_notify_event(GtkWidget *widget, GdkEventMotion *event)
{
  struct point_struc point, new_point;
  int x, y, grab = 1;
  GdkModifierType state;
  GdkCursorType new_cursor_type = cursor_type;
  GdkCursor *cursor;

  if (event->is_hint) {
    gdk_window_get_pointer(event->window, &x, &y, &state);
    point.x = x;
    point.y = y;
    if (!get_approx_point(widget, &grab_point, point.x, point.y))
      grab = 0;
  }
  else {
    point.x = event->x;
    point.y = event->y;
    state = event->state;
  }
  new_point = point;
  if (get_approx_point(widget, &new_point, point.x, point.y))
    new_cursor_type = GDK_HAND2;
  else new_cursor_type = GDK_TCROSS;
  if (new_cursor_type != cursor_type) {
    cursor_type = new_cursor_type;
    cursor = gdk_cursor_new (cursor_type);
    gdk_window_set_cursor (gtk_widget_get_parent_window(widget), cursor);
    gdk_cursor_destroy (cursor);
  }
  if (state & GDK_BUTTON1_MASK && pixmap != NULL) {
    if (grab) {
      new_point = point;
      unproject(widget->allocation.width, widget->allocation.height, &point);
      unproject
	(widget->allocation.width, widget->allocation.height, &grab_point);
      b_spline_update_control_poly_point(b_spline, grab_point, point);
      grab_point = new_point;
      redraw_scene(widget);
    }
  }
  return TRUE;
}

void quit()
{
  gtk_exit(0);
}

/* Draw control points of the curve */
void draw_control_points(GtkWidget* widget)
{
  int n, i;
  struct point_struc *points;
  GdkRectangle update_rect;

  update_rect.x = 0;
  update_rect.y = 0;
  update_rect.width = widget->allocation.width;
  update_rect.height = widget->allocation.height;

  b_spline_get_control_poly_points(b_spline, &n, &points);
  for (i = 0; i < n; ++i) {
    project(widget->allocation.width, widget->allocation.height, &points[i]);
    gdk_draw_rectangle (pixmap, widget->style->black_gc, TRUE,
			points[i].x - control_point_size / 2,
			points[i].y - control_point_size / 2,
			control_point_size, control_point_size);
  }
}

/* Draw curve on the screen */
static void draw_curve(GtkWidget *widget)
{
  GdkRectangle update_rect;
  float u;
  struct point_struc point;
  GdkGC *my_gc;
  GdkColor my_color;

  update_rect.x = 0;
  update_rect.y = 0;
  update_rect.width = widget->allocation.width;
  update_rect.height = widget->allocation.height;

  my_gc = gdk_gc_new(gtk_widget_get_parent_window(widget));
  my_color.red = 0; my_color.green = 0; my_color.blue = 65535; 
  gdk_colormap_alloc_color(gdk_colormap_get_system(), &my_color, TRUE, TRUE);
  gdk_gc_set_foreground (my_gc, &my_color);

  for (u = 0; u < b_spline_get_control_poly_points_count(b_spline);
       u += 0.01) {
    point = b_spline_deBoor_alg(b_spline, u);
    project(widget->allocation.width, widget->allocation.height, &point);
    gdk_draw_point (pixmap, my_gc, point.x, point.y);
  }
  gdk_gc_destroy(my_gc);
}

/* Draw control polygon on the screen */
static void draw_control_poly(GtkWidget *widget)
{
  GdkRectangle update_rect;
  int n, i;
  struct point_struc point, *points;
  GdkGC *my_gc;
  GdkColor my_color;

  update_rect.x = 0;
  update_rect.y = 0;
  update_rect.width = widget->allocation.width;
  update_rect.height = widget->allocation.height;

  my_gc = gdk_gc_new(gtk_widget_get_parent_window(widget));
  my_color.red = 65535; my_color.green = 0; my_color.blue = 0; 
  gdk_colormap_alloc_color(gdk_colormap_get_system(), &my_color, TRUE, TRUE);
  gdk_gc_set_foreground (my_gc, &my_color);

  b_spline_get_control_poly_points(b_spline, &n, &points);
  for (i = 0; i < n; ++i) {
    project(widget->allocation.width, widget->allocation.height, &points[i]);
    if (i == 0) { point = points[i]; continue; }
    gdk_draw_line(pixmap, my_gc, point.x, point.y, points[i].x, points[i].y);
    point = points[i];
  }
  gdk_gc_destroy(my_gc);
}

static gint lp_button_click_event(GtkWidget *widget)
{
  b_spline_set_level(b_spline, b_spline_get_level(b_spline) + 1);
  redraw_scene(widget);

  return TRUE;
}

static gint lm_button_click_event(GtkWidget *widget)
{
  b_spline_set_level(b_spline, b_spline_get_level(b_spline) - 1);
  redraw_scene(widget);

  return TRUE;
}

static gint check_button_toggle_event(GtkWidget *widget)
{
  show_control_poly = !show_control_poly;
  redraw_scene(widget);

  return TRUE;
}

int main(int argc, char *argv[])
{
  GtkWidget *window;
  GtkWidget *drawing_area;
  GtkWidget *vbox, *hbox1, *hbox2, *hbox3;
  GtkWidget *quit_button, *lp_button, *lm_button;
  GtkWidget *check_button;
  GtkWidget *pv_label;

  b_spline = b_spline_new();
  
  gtk_set_locale ();
  gtk_init (&argc, &argv);

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), "B-spline");

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  gtk_widget_show (vbox);

  hbox1 = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (vbox), hbox1, FALSE, FALSE, 0);
  gtk_widget_show (hbox1);

  stat_label = gtk_label_new("Level: ");
  gtk_box_pack_start (GTK_BOX (hbox1), stat_label, FALSE, FALSE, 2);
  gtk_widget_show (stat_label);

  check_button = gtk_check_button_new_with_label("Show control poly");
  gtk_box_pack_end (GTK_BOX (hbox1), check_button, FALSE, FALSE, 2);
  gtk_widget_show (check_button);

  hbox2 = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (vbox), hbox2, FALSE, FALSE, 0);
  gtk_widget_show (hbox2);

  pv_label = gtk_label_new("Param vector: ");
  gtk_box_pack_start (GTK_BOX (hbox2), pv_label, FALSE, FALSE, 2);
  gtk_widget_show (pv_label);

  entry = gtk_entry_new();
  gtk_box_pack_start (GTK_BOX (hbox2), entry, TRUE, TRUE, 0);
  gtk_widget_show (entry);

  gtk_signal_connect (GTK_OBJECT (window), "destroy",
		      GTK_SIGNAL_FUNC (quit), NULL);

  /* Create the drawing area */

  drawing_area = gtk_drawing_area_new ();
  gtk_drawing_area_size (GTK_DRAWING_AREA (drawing_area), 100, 100);
  gtk_box_pack_start (GTK_BOX (vbox), drawing_area, TRUE, TRUE, 0);
  gtk_widget_show (drawing_area);

  gtk_signal_connect_object (GTK_OBJECT (check_button), "toggled",
			     (GtkSignalFunc) check_button_toggle_event, 
			     GTK_OBJECT (drawing_area));
  gtk_signal_connect_object (GTK_OBJECT (entry), "activate",
			     (GtkSignalFunc) set_param_vector, 
			     GTK_OBJECT (drawing_area));

  /* Signals used to handle backing pixmap */

  gtk_signal_connect (GTK_OBJECT (drawing_area), "expose_event",
		      (GtkSignalFunc) expose_event, NULL);
  gtk_signal_connect (GTK_OBJECT(drawing_area),"configure_event",
		      (GtkSignalFunc) configure_event, NULL);

  /* Event signals */

  gtk_signal_connect (GTK_OBJECT (drawing_area), "motion_notify_event",
		      (GtkSignalFunc) motion_notify_event, NULL);
  gtk_signal_connect (GTK_OBJECT (drawing_area), "button_press_event",
		      (GtkSignalFunc) button_press_event, NULL);

  gtk_widget_set_events (drawing_area, GDK_EXPOSURE_MASK
			 | GDK_LEAVE_NOTIFY_MASK
			 | GDK_BUTTON_PRESS_MASK
			 | GDK_POINTER_MOTION_MASK
			 | GDK_POINTER_MOTION_HINT_MASK);

  hbox3 = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (vbox), hbox3, FALSE, FALSE, 0);
  gtk_widget_show (hbox3);

  lm_button = gtk_button_new_with_label ("Level -");
  gtk_box_pack_start (GTK_BOX (hbox3), lm_button, FALSE, FALSE, 0);
  gtk_signal_connect_object (GTK_OBJECT (lm_button), "clicked",
			     (GtkSignalFunc) lm_button_click_event, 
			     GTK_OBJECT (drawing_area));
  gtk_widget_show (lm_button);

  lp_button = gtk_button_new_with_label ("Level +");
  gtk_box_pack_start (GTK_BOX (hbox3), lp_button, FALSE, FALSE, 0);
  gtk_signal_connect_object (GTK_OBJECT (lp_button), "clicked",
			     (GtkSignalFunc) lp_button_click_event, 
			     GTK_OBJECT (drawing_area));
  gtk_widget_show (lp_button);

  /* .. And a quit button */
  quit_button = gtk_button_new_with_label ("Quit");
  gtk_box_pack_start (GTK_BOX (hbox3), quit_button, TRUE, TRUE, 0);
  gtk_signal_connect_object (GTK_OBJECT (quit_button), "clicked",
			     GTK_SIGNAL_FUNC (gtk_widget_destroy),
			     GTK_OBJECT (window));
  gtk_widget_show (quit_button);

  gtk_widget_show (window);

  gtk_main ();

  return 0;
}
